@@ -1,9 +1,9 @@
 /*
  * pg_filedump.c - PostgreSQL file dump utility for dumping and
- *                 formatting heap(data), index and control files.
- *                 Version 8.2.0 for PostgreSQL 8.2
+ *                 formatting heap (data), index and control files.
+ *                 Version 8.3.0 for PostgreSQL 8.3
  *
- * Copyright (c) 2002-2007 Red Hat, Inc. All rights reserved.
+ * Copyright (c) 2002-2008 Red Hat, Inc. All rights reserved.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -63,7 +63,7 @@ DisplayOptions (unsigned int validOptions)
 {
   if (validOptions == OPT_RC_COPYRIGHT)
     printf
-      ("\nVersion 8.2.0 (PostgreSQL 8.2)  Copyright (c) 2002-2007 Red Hat, Inc.\n");
+      ("\nVersion 8.3.0 (PostgreSQL 8.3)  Copyright (c) 2002-2008 Red Hat, Inc.\n");
 
   printf
     ("\nUsage: pg_filedump [-abcdfhixy] [-R startblock [endblock]] [-S blocksize] file\n\n"
@@ -346,14 +346,14 @@ ConsumeOptions (int numOptions, char **options)
 	      blockOptions = itemOptions = 0;
 	    }
 	}
-      // The user has request a binary block dump... only -R and
+      // The user has requested a binary block dump... only -R and
       // -f are honoured
       else if (blockOptions & BLOCK_BINARY)
 	{
 	  blockOptions &= (BLOCK_BINARY | BLOCK_RANGE | BLOCK_FORCED);
 	  itemOptions = 0;
 	}
-      // The user has requested a non-intepreted dump... only -a,
+      // The user has requested a non-interpreted dump... only -a,
       // -R and -f are honoured
       else if (blockOptions & BLOCK_NO_INTR)
 	{
@@ -415,7 +415,7 @@ GetBlockSize ()
 static unsigned int
 GetSpecialSectionType (Page page)
 {
-  unsigned int rc = 0;
+  unsigned int rc;
   unsigned int specialOffset;
   unsigned int specialSize;
   unsigned int specialValue;
@@ -458,21 +458,24 @@ GetSpecialSectionType (Page page)
 	    }
 	  else if (specialSize == MAXALIGN (sizeof (GinPageOpaqueData)))
 	      rc = SPEC_SECT_INDEX_GIN;
-	  else if (specialSize == MAXALIGN (sizeof (HashPageOpaqueData)))
+	  else if (specialSize > 2 && bytesToFormat == blockSize)
 	    {
-	      // As of 7.4, BTree and Hash pages have the same size special
-	      // section.  Check for HASHO_FILL to detect if it's hash.
+	      // As of 8.3, BTree, Hash, and GIST all have the same size
+	      // special section, but the last two bytes of the section
+	      // can be checked to determine what's what.
+	      uint16 ptype = *(uint16 *) (buffer + blockSize - sizeof(uint16));
 
-	      // As of 8.2, it could be GIST, too ... but there seems no
-	      // good way to tell GIST from BTree :-(  Also, HASHO_FILL is
-	      // not reliable anymore, it could match cycleid by chance.
-	      // Need to try to get some upstream changes to make this better.
-
-	      HashPageOpaque hpo = (HashPageOpaque) (buffer + specialOffset);
-	      if (hpo->hasho_filler == HASHO_FILL)
-		rc = SPEC_SECT_INDEX_HASH;
-	      else
+	      if (ptype <= MAX_BT_CYCLE_ID &&
+		  specialSize == MAXALIGN (sizeof (BTPageOpaqueData)))
 		rc = SPEC_SECT_INDEX_BTREE;
+	      else if (ptype == HASHO_PAGE_ID &&
+		  specialSize == MAXALIGN (sizeof (HashPageOpaqueData)))
+		rc = SPEC_SECT_INDEX_HASH;
+	      else if (ptype == GIST_PAGE_ID &&
+		       specialSize == MAXALIGN (sizeof (GISTPageOpaqueData)))
+		rc = SPEC_SECT_INDEX_GIST;
+	      else
+		rc = SPEC_SECT_ERROR_UNKNOWN;
 	    }
 	  else
 	    rc = SPEC_SECT_ERROR_UNKNOWN;
@@ -493,12 +496,11 @@ IsBtreeMetaPage(Page page)
   if ((PageGetSpecialSize (page) == (MAXALIGN (sizeof (BTPageOpaqueData))))
       && (bytesToFormat == blockSize))
     {
-      // As of 7.4, BTree and Hash pages have the same size special
-      // section.  Check for HASHO_FILL to detect if it's hash.
       BTPageOpaque btpo =
 	(BTPageOpaque) ((char *) page + pageHeader->pd_special);
 
-      if ((((HashPageOpaque) (btpo))->hasho_filler != HASHO_FILL) &&
+      // Must check the cycleid to be sure it's really btree.
+      if ((btpo->btpo_cycleid <= MAX_BT_CYCLE_ID) &&
 	  (btpo->btpo_flags & BTP_META))
 	return true;
     }
@@ -526,7 +528,7 @@ CreateDumpFileHeader (int numOptions, char **options)
 
   printf
     ("\n*******************************************************************\n"
-     "* PostgreSQL File/Block Formatted Dump Utility - Version 8.2.0\n*\n"
+     "* PostgreSQL File/Block Formatted Dump Utility - Version 8.3.0\n*\n"
      "* File: %s\n"
      "* Options used: %s\n*\n"
      "* Dump created on: %s"
@@ -556,6 +558,8 @@ FormatHeader (Page page)
     {
       XLogRecPtr pageLSN = PageGetLSN (page);
       int maxOffset = PageGetMaxOffsetNumber (page);
+      char flagString[100];
+
       headerBytes = offsetof (PageHeaderData, pd_linp[0]);
       blockVersion = (unsigned int) PageGetPageLayoutVersion (page);
 
@@ -573,19 +577,31 @@ FormatHeader (Page page)
 	    headerBytes += itemsLength;
 	}
 
+      flagString[0] = '\0';
+      if (pageHeader->pd_flags & PD_HAS_FREE_LINES)
+	  strcat (flagString, "HAS_FREE_LINES|");
+      if (pageHeader->pd_flags & PD_PAGE_FULL)
+	  strcat (flagString, "PAGE_FULL|");
+      if (strlen (flagString))
+	  flagString[strlen (flagString) - 1] = '\0';
+
       // Interpret the content of the header
       printf
 	(" Block Offset: 0x%08x         Offsets: Lower    %4u (0x%04hx)\n"
 	 " Block: Size %4d  Version %4u            Upper    %4u (0x%04hx)\n"
 	 " LSN:  logid %6d recoff 0x%08x      Special  %4u (0x%04hx)\n"
-	 " Items: %4d                   Free Space: %4u\n"
+	 " Items: %4d                      Free Space: %4u\n"
+	 " TLI: 0x%04x  Prune XID: 0x%08x  Flags: 0x%04x (%s)\n"
 	 " Length (including item array): %u\n\n",
 	 pageOffset, pageHeader->pd_lower, pageHeader->pd_lower,
-	 (int) PageGetPageSize (page), blockVersion, pageHeader->pd_upper,
-	 pageHeader->pd_upper,
-	 pageLSN.xlogid, pageLSN.xrecoff, pageHeader->pd_special,
-	 pageHeader->pd_special, maxOffset,
-	 pageHeader->pd_upper - pageHeader->pd_lower, headerBytes);
+	 (int) PageGetPageSize (page), blockVersion,
+	 pageHeader->pd_upper, pageHeader->pd_upper,
+	 pageLSN.xlogid, pageLSN.xrecoff,
+	 pageHeader->pd_special, pageHeader->pd_special,
+	 maxOffset, pageHeader->pd_upper - pageHeader->pd_lower,
+	 pageHeader->pd_tli, pageHeader->pd_prune_xid,
+	 pageHeader->pd_flags, flagString,
+	 headerBytes);
 
       // If it's a btree meta page, print the contents of the meta block.
       if (IsBtreeMetaPage(page))
@@ -660,7 +676,7 @@ FormatItemBlock (Page page)
   else
     {
       int formatAs;
-      char textFlags[8];
+      char textFlags[16];
 
       // First, honour requests to format items a special way, then 
       // use the special section to determine the format style
@@ -680,10 +696,25 @@ FormatItemBlock (Page page)
 	  itemSize = (unsigned int) ItemIdGetLength (itemId);
 	  itemOffset = (unsigned int) ItemIdGetOffset (itemId);
 
-	  if (itemFlags)
-	    strcpy (textFlags, (LP_USED & itemFlags) ? "USED" : "DELETE");
-	  else
-	    sprintf (textFlags, "0x%02x", itemFlags);
+	  switch (itemFlags)
+	  {
+	    case LP_UNUSED:
+	      strcpy (textFlags, "UNUSED");
+	      break;
+	    case LP_NORMAL:
+	      strcpy (textFlags, "NORMAL");
+	      break;
+	    case LP_REDIRECT:
+	      strcpy (textFlags, "REDIRECT");
+	      break;
+	    case LP_DEAD:
+	      strcpy (textFlags, "DEAD");
+	      break;
+	    default:
+	      // shouldn't be possible
+	      sprintf (textFlags, "0x%02x", itemFlags);
+	      break;
+	  }
 
 	  printf (" Item %3u -- Length: %4u  Offset: %4u (0x%04x)"
 		  "  Flags: %s\n", x, itemSize, itemOffset, itemOffset,
@@ -761,6 +792,7 @@ FormatItem (unsigned int numBytes, unsigned int startIndex,
 	  unsigned int oidLength = 0;
 	  unsigned int computedLength;
 	  unsigned int infoMask;
+	  unsigned int infoMask2;
 	  int localNatts;
 	  unsigned int localHoff;
 	  bits8 *localBits;
@@ -769,16 +801,16 @@ FormatItem (unsigned int numBytes, unsigned int startIndex,
 	  HeapTupleHeader htup = (HeapTupleHeader) (&buffer[startIndex]);
 
 	  infoMask = htup->t_infomask;
+	  infoMask2 = htup->t_infomask2;
 	  localBits = &(htup->t_bits[0]);
-	  localNatts = htup->t_natts;
+	  localNatts = HeapTupleHeaderGetNatts(htup);
 	  localHoff = htup->t_hoff;
 	  localBitOffset = offsetof (HeapTupleHeaderData, t_bits);
 
-	  printf ("  XMIN: %u  CMIN: %u  XMAX: %u  CMAX|XVAC: %u",
+	  printf ("  XMIN: %u  XMAX: %u  CID|XVAC: %u",
 		  HeapTupleHeaderGetXmin(htup),
-		  HeapTupleHeaderGetCmin(htup),
 		  HeapTupleHeaderGetXmax(htup),
-		  HeapTupleHeaderGetCmax(htup));
+		  HeapTupleHeaderGetRawCommandId(htup));
 
 	  if (infoMask & HEAP_HASOID)
 	    printf ("  OID: %u",
@@ -789,7 +821,7 @@ FormatItem (unsigned int numBytes, unsigned int startIndex,
 		  ((uint32)
 		   ((htup->t_ctid.ip_blkid.bi_hi << 16) | (uint16) htup->
 		    t_ctid.ip_blkid.bi_lo)), htup->t_ctid.ip_posid,
-		  htup->t_natts, htup->t_hoff);
+		  localNatts, htup->t_hoff);
 
 	  // Place readable versions of the tuple info mask into a buffer.  
 	  // Assume that the string can not expand beyond 256.
@@ -800,10 +832,10 @@ FormatItem (unsigned int numBytes, unsigned int startIndex,
 	    strcat (flagString, "HASVARWIDTH|");
 	  if (infoMask & HEAP_HASEXTERNAL)
 	    strcat (flagString, "HASEXTERNAL|");
-	  if (infoMask & HEAP_HASCOMPRESSED)
-	    strcat (flagString, "HASCOMPRESSED|");
 	  if (infoMask & HEAP_HASOID)
 	    strcat (flagString, "HASOID|");
+	  if (infoMask & HEAP_COMBOCID)
+	    strcat (flagString, "COMBOCID|");
 	  if (infoMask & HEAP_XMAX_EXCL_LOCK)
 	    strcat (flagString, "XMAX_EXCL_LOCK|");
 	  if (infoMask & HEAP_XMAX_SHARED_LOCK)
@@ -824,6 +856,12 @@ FormatItem (unsigned int numBytes, unsigned int startIndex,
 	    strcat (flagString, "MOVED_OFF|");
 	  if (infoMask & HEAP_MOVED_IN)
 	    strcat (flagString, "MOVED_IN|");
+
+	  if (infoMask2 & HEAP_HOT_UPDATED)
+	    strcat (flagString, "HOT_UPDATED|");
+	  if (infoMask2 & HEAP_ONLY_TUPLE)
+	    strcat (flagString, "HEAP_ONLY|");
+
 	  if (strlen (flagString))
 	    flagString[strlen (flagString) - 1] = '\0';
 
@@ -958,14 +996,16 @@ FormatSpecial ()
 	if (gistSection->flags & F_DELETED)
 	  strcat (flagString, "DELETED|");
 	if (gistSection->flags & F_TUPLES_DELETED)
-	  strcat (flagString, "TUPLESDELETED|");
+	  strcat (flagString, "TUPLES_DELETED|");
 	if (strlen (flagString))
 	  flagString[strlen (flagString) - 1] = '\0';
 	printf (" GIST Index Section:\n"
-		"  Flags: 0x%08x (%s)\n"
-		"  Blocks: RightLink (%d)\n\n",
-		gistSection->flags, flagString,
-		gistSection->rightlink);
+		"  NSN: 0x%08x/0x%08x\n"
+		"  RightLink: %d\n"
+		"  Flags: 0x%08x (%s)\n\n",
+		gistSection->nsn.xlogid, gistSection->nsn.xrecoff,
+		gistSection->rightlink,
+		gistSection->flags, flagString);
       }
       break;
 
@@ -1060,7 +1100,8 @@ FormatControl ()
     controlFileSize = sizeof (ControlFileData);
   else
     {
-      printf ("pg_filedump: PostgreSQL %u not supported.\n", localPgVersion);
+      printf ("pg_filedump: pg_control version %u not supported.\n",
+	      localPgVersion);
       return;
     }
 
@@ -1108,26 +1149,26 @@ FormatControl ()
 	      "              Catalog Version: %u\n"
 	      "            System Identifier: " UINT64_FORMAT "\n"
 	      "                        State: %s\n"
-	      "              Last Checkpoint: %s"
-	      "             Current Log File: %u\n"
-	      "             Next Log Segment: %u\n"
+	      "                Last Mod Time: %s"
 	      "       Last Checkpoint Record: Log File (%u) Offset (0x%08x)\n"
 	      "   Previous Checkpoint Record: Log File (%u) Offset (0x%08x)\n"
 	      "  Last Checkpoint Record Redo: Log File (%u) Offset (0x%08x)\n"
-	      "             |-          Undo: Log File (%u) Offset (0x%08x)\n"
 	      "             |-    TimeLineID: %u\n"
-	      "             |-      Next XID: %u\n"
+	      "             |-      Next XID: %u/%u\n"
 	      "             |-      Next OID: %u\n"
 	      "             |-    Next Multi: %u\n"
 	      "             |- Next MultiOff: %u\n"
 	      "             |-          Time: %s"
+	      "       Minimum Recovery Point: Log File (%u) Offset (0x%08x)\n"
 	      "       Maximum Data Alignment: %u\n"
 	      "        Floating-Point Sample: %.7g%s\n"
 	      "          Database Block Size: %u\n"
 	      "           Blocks Per Segment: %u\n"
+	      "              XLOG Block Size: %u\n"
 	      "            XLOG Segment Size: %u\n"
 	      "    Maximum Identifier Length: %u\n"
 	      "           Maximum Index Keys: %u\n"
+	      "             TOAST Chunk Size: %u\n"
 	      "   Date and Time Type Storage: %s\n"
 	      "         Locale Buffer Length: %u\n"
 	      "                   lc_collate: %s\n"
@@ -1140,28 +1181,31 @@ FormatControl ()
 	      controlData->catalog_version_no,
 	      controlData->system_identifier,
 	      dbState,
-	      ctime (&(controlData->time)), controlData->logId,
-	      controlData->logSeg, controlData->checkPoint.xlogid,
-	      controlData->checkPoint.xrecoff,
-	      controlData->prevCheckPoint.xlogid,
-	      controlData->prevCheckPoint.xrecoff, checkPoint->redo.xlogid,
-	      checkPoint->redo.xrecoff, checkPoint->undo.xlogid,
-	      checkPoint->undo.xrecoff, checkPoint->ThisTimeLineID,
-	      checkPoint->nextXid, checkPoint->nextOid,
+	      ctime (&(controlData->time)),
+	      controlData->checkPoint.xlogid, controlData->checkPoint.xrecoff,
+	      controlData->prevCheckPoint.xlogid, controlData->prevCheckPoint.xrecoff,
+	      checkPoint->redo.xlogid, checkPoint->redo.xrecoff,
+	      checkPoint->ThisTimeLineID,
+	      checkPoint->nextXidEpoch, checkPoint->nextXid,
+	      checkPoint->nextOid,
 	      checkPoint->nextMulti, checkPoint->nextMultiOffset,
 	      ctime (&checkPoint->time),
+	      controlData->minRecoveryPoint.xlogid, controlData->minRecoveryPoint.xrecoff,
 	      controlData->maxAlign,
 	      controlData->floatFormat,
 	      (controlData->floatFormat == FLOATFORMAT_VALUE ?
 	       "" : " (Not Correct!)"),
 	      controlData->blcksz,
 	      controlData->relseg_size,
+	      controlData->xlog_blcksz,
 	      controlData->xlog_seg_size,
 	      controlData->nameDataLen,
 	      controlData->indexMaxKeys,
+	      controlData->toast_max_chunk_size,
 	      (controlData->enableIntTimes ?
 	       "64 bit Integers" : "Floating Point"),
-	      controlData->localeBuflen, controlData->lc_collate,
+	      controlData->localeBuflen,
+	      controlData->lc_collate,
 	      controlData->lc_ctype);
     }
   else
