@@ -1,9 +1,9 @@
 /*
  * pg_filedump.h - PostgreSQL file dump utility for dumping and
- *                 formatting heap(data), index and control files.
- *                 Version 8.2.0 for PostgreSQL 8.2
+ *                 formatting heap (data), index and control files.
+ *                 Version 8.3.0 for PostgreSQL 8.3
  * 
- * Copyright (c) 2002-2007 Red Hat, Inc. All rights reserved.
+ * Copyright (c) 2002-2008 Red Hat, Inc. All rights reserved.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -60,8 +60,8 @@ static unsigned int itemOptions = 0;
 typedef enum
 {
   ITEM_DETAIL = 0x00000001,	// -i: Display interpreted items    
-  ITEM_HEAP = 0x00000002,	// -x: Blocks contain heap items
-  ITEM_INDEX = 0x00000004	// -y: Blocks contain heap items
+  ITEM_HEAP = 0x00000002,	// -y: Blocks contain heap items
+  ITEM_INDEX = 0x00000004	// -x: Blocks contain index items
 }
 itemSwitches;
 
